#include <stdio.h>
#include "tree.h"
#include "symbol_table.h"
#include "prover.h"

typedef struct {
  int operator;
  unsigned int value_number;
} formula_key;

typedef struct formula_hash {
  formula_key key;
  formulalist *list;
  UT_hash_handle hf;
} formula_hash;

extern int numprops;
extern int newtemp;
extern int ires;
extern int ml_prover;
extern int numclause;
extern int global2local;
extern int mdepth;

extern int is_literal (tnode *);
extern int same_tree(tnode *t1, tnode *t2);
extern tnode *copy_tree(tnode *s);
extern tnode *free_tree (tnode *s);
extern void print_tree (tnode *s);
extern tnode *tree_hash(tnode *node);
extern void create_clauses (int set, tnode *s);
extern literalslist *simplify_disjunction(literal_list *l);
extern int size_clause (literalslist *l);
extern int max_lit_clause (literalslist *l);
extern int classify_clause (literalslist *l);
extern void insert_clause_set (tclause **clause, int type, int set);
extern prop_node *insert_pnew_node (int id, int type);
extern tnode *create_tnode(int type, int id, int mdepth, tnode *left, tnode *right, formulalist *list);
extern void create_clause_snf(int set, int type, int subtype, int modal_level, int agent, int numdiamond, tnode *left, literal_list *right, justification *just);
extern void print_clause(tclause *clause);
extern void copy_node(tnode *t1, tnode *t2);
extern agent_node *find_agent (int id);   
formula_hash *fd = NULL;

int numcolisions = 0;
int numsubformula = 0;
int sizeformula = 0;
int numliterals = 0;
int numboxes = 0;
int numdiamonds = 0;


void create_formula_dict(tnode *formula) {
  if (formula != NULL && (formula->type == SETF || formula->type == SETC))
    create_formula_dict(formula->left);
  else if (formula != NULL && is_literal(formula)) {
    numliterals++;
  }
  else if (formula != NULL) {
    if (formula->type == BOX) numboxes++;
    else if (formula->type == DIAMOND) numdiamonds++;
    
    formula_hash *entry = NULL;
    formula_key key;
    
    key.operator = formula->type;
    key.value_number = formula->value_number;

    HASH_FIND(hf,fd,&(key),sizeof(formula_key),entry);

    if (entry != NULL) { // there is an entry in the dictionary, need to check if the formula is already there
      formulalist *auxl = entry->list;
      while (auxl != NULL && !same_tree(formula,auxl->formula))
	auxl = auxl->next;
      if (auxl == NULL) { // not in the dictionary yet
	formulalist *new_list = malloc(sizeof(formulalist));      
	new_list->mdepth = formula->mdepth;
	new_list->value_number = formula->value_number;
	prop_node *p = insert_pnew_node(newtemp++,HIGH);
	new_list->renamed_by = p->id;
	new_list->definition_added = 0;
	new_list->formula = copy_tree(formula);
	//	new_list->formula = tree_hash(new_list->formula);
	new_list->next = entry->list;
	entry->list = new_list;
	numsubformula++;
      }
      else numcolisions++;
    }
    else { // there is no entry in the dictionary, need to create one.

      	formula_hash *new = malloc(sizeof(formula_hash));
	new->list = NULL;
    
	formula_key new_key;
	new_key.operator = formula->type;
	new_key.value_number = formula->value_number;

	new->key = new_key;
	
	formulalist *new_list = malloc(sizeof(formulalist));      
	new_list->mdepth = formula->mdepth;
	new_list->value_number = formula->value_number;
	prop_node *p = insert_pnew_node(newtemp++,HIGH);
	new_list->renamed_by = p->id;
	new_list->definition_added = 0;
	new_list->formula = copy_tree(formula);
	//	new_list->formula = tree_hash(new_list->formula);
	new_list->next = NULL;

	new->list = new_list;
	
	HASH_ADD(hf,fd,key,sizeof(formula_key),new);
	numsubformula++;
    }
	
    create_formula_dict(formula->left);
    create_formula_dict(formula->right);
    formulalist *auxl = formula->list;
    while (auxl != NULL) {
      create_formula_dict(auxl->formula);
      auxl = auxl->next;
    }
  }
}


void print_formula_dict(formula_hash *dict) {
  formula_hash *tmp,*tmpp;
  printf("\n FORMULA TABLE\n");
  HASH_ITER(hf,dict,tmp,tmpp) {
    printf("\n NEW ENTRY: Operator: %d, Value_number: %u",tmp->key.operator,tmp->key.value_number);
    formulalist *aux = tmp->list;
    while (aux != NULL) {
      printf("\n Renamed by: %d\n",aux->renamed_by);
      print_tree(aux->formula); 
      aux = aux->next;
    }
  }
}


void create_dict(tnode *root) {
  root = tree_hash(root);
  create_formula_dict(root);
}


int new_renaming (tnode *formula) {
   formula_hash *entry = NULL;
   formula_key key;
    
   key.operator = formula->type;
   key.value_number = formula->value_number;

   HASH_FIND(hf,fd,&(key),sizeof(formula_key),entry);
   if (entry == NULL) {
     return 0;
   }
   else {
     formulalist *aux = entry->list;
     while (aux != NULL) {
       if (same_tree(formula,aux->formula)) return aux->renamed_by;
       aux = aux->next;
     }
   }
   return 0;
}



tnode *negate_literal(tnode *s) {
  if (s->type == NEGATION) return s->left;
  else return create_tnode(NEGATION,NEGATION,0,s,NULL,NULL);
}

tnode *create_tnode_lit(int id){
  tnode *new = create_tnode(PROP,id,0,NULL,NULL,NULL);
  if (id < 0) {
    new->id = -new->id;
    tnode *neg = create_tnode(NEGATION,NEGATION,0,new,NULL,NULL);
    new = neg;
  }
  return new;
}

void create_clause_snf_b(int set, int type, int subtype, int modal_level, int agent, int numdiamond, tnode *left, formulalist  *right, justification *just) {

  tclause *clause = malloc(sizeof(tclause));
  clause->number = numclause++;
  clause->type = type;
  
  /*  if (type == INITIAL)
    clause->subtype = BOX;
    else */
  clause->subtype = subtype;
  clause->numdiamond = numdiamond;
  
  if (!ml_prover)
    clause->modal_level = 1;
  else
    clause->modal_level = modal_level;
  clause->agent = agent;

  literal_list *dis = NULL;

  if (type == INITIAL) clause->left = CSTART;
  else if (type == LITERAL) clause->left = CTRUE;
  else if (left->type == NEGATION) clause->left = -left->left->id;
  else clause->left = left->id;

  if (left->type != CONSTANT && (type == LITERAL || type == INITIAL)) {
    dis = malloc(sizeof(literal_list));
    dis->literal = negate_literal(left);
    dis->next = NULL;
  }
 
  formulalist *aux = right;

  while (aux != NULL) {
    literal_list *newdis = malloc(sizeof(literal_list));
    newdis->literal = aux->formula;
    newdis->next = dis;
    dis = newdis;
    aux = aux->next;
  }
    
  clause->right = simplify_disjunction(dis);
  clause->just = just;
  clause->deleted = NULL;
  clause->propagated = 0;
  clause->size = size_clause(clause->right);
  clause->max_literal = max_lit_clause(clause->right);
  clause->min_literal = clause->right->literal;
  clause->class = classify_clause(clause->right);
  insert_clause_set(&clause,clause->type,set);
}


/* The input formula is an implication */

int get_snf_b (int set, int type, int subtype, int mlevel, tnode *s, int recur) {

  //printf("\nFormula is:\n");
  //print_tree(s);
      
  if (s == NULL)
    return 0;
  if (s->type != IMPLICATION) {
    printf("\n SNF is taking argument of wrong type: %d \n", s->type);
  }
  else {
    switch (s->right->type) {
    case PROP:
    case CONSTANT:
    case NEGATION:
	{
	  tnode *left = copy_tree(s->left);
	  formulalist *aux = malloc(sizeof(formulalist));
	  aux->formula = copy_tree(s->right);
	  aux->next = NULL;

	  justification *just = malloc(sizeof(justification));
	  just->rule = recur;
	  just->parents = NULL;
	  just->literals = NULL;

	  create_clause_snf_b(set,LITERAL,subtype,mlevel,PROP,1,left,aux,just);
	}
      break;
    case CONJUNCTION:
      {
	formulalist *list = s->right->list;
	while (list != NULL) {
	  tnode *left = copy_tree(s->left);
	  tnode *right = copy_tree(list->formula);
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left,right,NULL);
	  get_snf_b(set,type,subtype,mlevel,newimp,recur);
	  list = list->next;
	}
      }
      break;
    case DISJUNCTION:
      {
	formulalist *right = s->right->list;
	formulalist *newlist = NULL;
	while (right != NULL) {
	  if (!(is_literal(right->formula))) {
	      int def = new_renaming(right->formula);
	      tnode *new_lit_1 = create_tnode_lit(def);
	      tnode *new_lit_2 = create_tnode_lit(def);
	      tnode *new_imp = create_tnode(IMPLICATION,IMPLICATION,s->right->mdepth,new_lit_1,right->formula,NULL);
	      get_snf_b(set,type,subtype,mlevel,new_imp,recur);
	      formulalist *aux = malloc(sizeof(formulalist));
	      aux->formula = new_lit_2;
	      aux->next = newlist;
	      newlist = aux;
	    }
	  else {
	    formulalist *aux = malloc(sizeof(formulalist));
	    aux->formula = copy_tree(right->formula);
	    aux->next = newlist;
	    newlist = aux;
	  }
	  right = right->next;
	}

	justification *just = malloc(sizeof(justification));
	just->rule = recur;
	just->parents = NULL;
	just->literals = NULL;

	tnode *left = copy_tree(s->left);
	
	create_clause_snf_b(set,LITERAL,subtype,mlevel,PROP,1,left,newlist,just);
      }
      break;
    case DIAMOND:
    case BOX:
      {
	int clause_type = MPOSITIVE;
	if (s->right->type == DIAMOND) clause_type = MNEGATIVE;
	int clause_subtype = s->right->numdiamond;

	tnode *new_right = NULL;

	agent_node *a = find_agent(s->right->id);
	
	if (!(is_literal(s->right->left))) {
	  int def = new_renaming(s->right->left);
	  //	  print_entry_dict(def,fd);
	  tnode *new_lit_1 = create_tnode_lit(def);
	  tnode *new_lit_2 = create_tnode_lit(def);
	  tnode *right = copy_tree(s->right->left);
	  tnode *new_imp = create_tnode(IMPLICATION,IMPLICATION,s->right->mdepth,new_lit_1,right,NULL);
	  if (ml_prover == ON) 
	    get_snf_b(set,clause_type,clause_subtype,mlevel+1,new_imp,recur);
	  else get_snf_b(set,clause_type,clause_subtype,1,new_imp,recur);
	  new_right = new_lit_2;
	}
	else new_right = copy_tree(s->right->left);
	
	justification *just = malloc(sizeof(justification));
	just->rule = recur;
	just->parents = NULL;
	just->literals = NULL;

	formulalist *list = malloc(sizeof(formulalist));
	list->formula = new_right;
	list->next = NULL;

	tnode *new_left = copy_tree(s->left);
	create_clause_snf_b(set,clause_type,clause_subtype,mlevel,s->right->id,s->left->numdiamond,new_left,list,just);

	if (clause_type == MPOSITIVE && ((a->axioms & T) != 0) && (recur != T)) {
	  if (!same_tree(s->left,s->right->left)) {
	    tnode *left = copy_tree(s->left);
	    tnode *right = copy_tree(s->right->left);
	    tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left,right,NULL);
	    get_snf_b(set,LITERAL,clause_subtype,mlevel,newimp,T);
	  }
	}

	if (clause_type == MPOSITIVE && ((a->axioms & SER) != 0) && (recur != SER)) {
	  tnode *left = copy_tree(s->left);
	  tnode *right = copy_tree(s->right->left);
	  tnode *dia = create_tnode(DIAMOND,s->right->id,s->mdepth,right,NULL,NULL);
	  create_formula_dict(dia);
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left,dia,NULL);
	  get_snf_b(set,MNEGATIVE,clause_subtype,mlevel,newimp,SER);
	}

	if (clause_type == MPOSITIVE && ((a->axioms & FOUR1) != 0) && (recur != FOUR1)) {
	  int levels = 0;
	  int def = new_renaming(s->right);

	  if (def != s->left->id) {
	    tnode *left3 = copy_tree(s->left);
	    tnode *neg = create_tnode(NEGATION,NEGATION,mlevel,left3,NULL,NULL);
	    tnode *right3 = create_tnode(PROP,def,mlevel,NULL,NULL,NULL);
	    formulalist *newlist = malloc(sizeof(formulalist));
	    newlist->formula = neg;
	    newlist->next = malloc(sizeof(formulalist));
	    newlist->next->formula = right3;
	    newlist->next->next = NULL;
	    tnode *left = create_tnode(CONSTANT,CTRUE,mlevel,NULL,NULL,NULL);
	    tnode *disj = create_tnode(DISJUNCTION,DISJUNCTION,mlevel,NULL,NULL,newlist);
	    tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,mlevel,left,disj,NULL);
	    get_snf_b(set,LITERAL,clause_subtype,mlevel,newimp,FOUR1);
	  }
	  	  
	  for (levels = mlevel; levels <= mdepth + 1; levels++) {
	    if (def != s->left->id) {
	      tnode *left1 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	      tnode *right1 = copy_tree(s->right);
	      tnode *newimp1 = create_tnode(IMPLICATION,IMPLICATION,levels,left1,right1,NULL);
	      get_snf_b(set,MPOSITIVE,clause_subtype,levels,newimp1,FOUR1);
	    }
	  	    
	    tnode *left2 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *right2 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *box = create_tnode(BOX,s->right->id,levels,right2,NULL,NULL);
	    tnode *newimp2 = create_tnode(IMPLICATION,IMPLICATION,levels,left2,box,NULL);
	    get_snf_b(set,MPOSITIVE,clause_subtype,levels,newimp2,FOUR1);
	   
	  if (!(is_literal(s->right->left))) {
	      int newdef = new_renaming(s->right->left);
	      if (newdef == 0) {
		if (s->right->left->type == NEGATION)  newdef = s->right->left->left->id;
		else newdef = s->right->left->id;
	      }
	      tnode *left4 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	      if (!(is_literal(s->right->left))) {
		tnode *right4 = copy_tree(s->right->left);
		tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,levels,left4,right4,NULL);
		get_snf_b(set,LITERAL,clause_subtype,levels,newimpred,FOUR1);
	      }
	    }
	  }
	}


	if (clause_type == MPOSITIVE && ((a->axioms & FIVE) != 0) && (recur != FIVE)) {
	  int levels = 0;
	  int def = new_renaming(s->right);

	  if (def != s->left->id) {
	    tnode *left3 = copy_tree(s->left);
	    tnode *neg = create_tnode(NEGATION,NEGATION,mlevel,left3,NULL,NULL);
	    tnode *right3 = create_tnode(PROP,def,mlevel,NULL,NULL,NULL);
	    formulalist *newlist = malloc(sizeof(formulalist));
	    newlist->formula = neg;
	    newlist->next = malloc(sizeof(formulalist));
	    newlist->next->formula = right3;
	    newlist->next->next = NULL;
	    tnode *left = create_tnode(CONSTANT,CTRUE,mlevel,NULL,NULL,NULL);
	    tnode *disj = create_tnode(DISJUNCTION,DISJUNCTION,mlevel,NULL,NULL,newlist);
	    tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,mlevel,left,disj,NULL);
	    get_snf_b(set,LITERAL,clause_subtype,mlevel,newimp,FOUR1);
	  }
	  
	  for (levels = 1; levels <= mdepth+1; levels++) {
	    tnode *left1 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *right1 = copy_tree(s->right);
	    tnode *newimp1 = create_tnode(IMPLICATION,IMPLICATION,levels,left1,right1,NULL);
	    get_snf_b(set,MPOSITIVE,clause_subtype,levels,newimp1,FIVE);
	      
	  	    
	    tnode *right2 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *dia = create_tnode(DIAMOND,s->right->id,levels,right2,NULL,NULL);
	    create_formula_dict(dia);
	    int newdef = new_renaming(dia);
	    tnode *left2 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	    
	    tnode *newimp2 = create_tnode(IMPLICATION,IMPLICATION,levels,left2,dia,NULL);
	    get_snf_b(set,MNEGATIVE,clause_subtype,levels,newimp2,FIVE);

	    tnode *right3 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *neg1 = create_tnode(NEGATION,NEGATION,levels,right3,NULL,NULL);
	    tnode *box = create_tnode(BOX,s->right->id,levels,neg1,NULL,NULL);
	    
	    tnode *left3 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	    tnode *neg2 = create_tnode(NEGATION,NEGATION,levels,left3,NULL,NULL);

	    tnode *newimp3 = create_tnode(IMPLICATION,IMPLICATION,levels,neg2,box,NULL);
	    get_snf_b(set,MPOSITIVE,clause_subtype,levels,newimp3,FIVE);


	    tnode *left4 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	    tnode *right4 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	    tnode *box2 = create_tnode(BOX,s->right->id,levels,right4,NULL,NULL);
	    tnode *newimp4 = create_tnode(IMPLICATION,IMPLICATION,levels,left4,box2,NULL);
	    get_snf_b(set,MPOSITIVE,clause_subtype,levels,newimp4,FIVE);

	    tnode *left5 = create_tnode(PROP,newdef,levels,NULL,NULL,NULL);
	    tnode *right5 = create_tnode(PROP,def,levels,NULL,NULL,NULL);
	    tnode *newimp5 = create_tnode(IMPLICATION,IMPLICATION,levels,left5,right5,NULL);
	    get_snf_b(set,LITERAL,clause_subtype,levels,newimp5,FIVE);
	    
	  if (!(is_literal(s->right->left))) {
	      int newdef2 = new_renaming(s->right->left);
	      if (newdef2 == 0) {
		if (s->right->left->type == NEGATION)  newdef2 = s->right->left->left->id;
		else newdef2 = s->right->left->id;
	      }
	      tnode *left6 = create_tnode(PROP,newdef2,levels,NULL,NULL,NULL);
	      if (!(is_literal(s->right->left))) {
		tnode *right6 = copy_tree(s->right->left);
		tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,levels,left6,right6,NULL);
		get_snf_b(set,LITERAL,clause_subtype,levels,newimpred,FIVE);
	      }
	    }
	  }
	}

	
	if (mlevel != 1 && clause_type == MPOSITIVE && ((a->axioms & SYM) != 0) && (recur != SYM)) {
	  /* t_([a]~t) => [a]~t : in the predecessor */
	  tnode *left1 = copy_tree(s->left);
	  tnode *neg = create_tnode(NEGATION,NEGATION,s->mdepth,left1,NULL,NULL);
	  tnode *box = create_tnode(BOX,s->right->id,s->mdepth,neg,NULL,NULL);
	  create_formula_dict(box);
	  int def = new_renaming(box);
	  tnode *propdef = create_tnode(PROP,def,s->mdepth,NULL,NULL,NULL);
	  tnode *newimpdef = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,propdef,box,NULL);
	  get_snf_b(set,MPOSITIVE,clause_subtype,mlevel-1,newimpdef,SYM);

	  /* t_phi \/ t_([a]~t) : in the predecessor */

	  tnode *left2 = copy_tree(s->right->left);
	  tnode *right2 = copy_tree(propdef);
	  formulalist *newlist = malloc(sizeof(formulalist));
	  newlist->formula = left2;
	  newlist->next = malloc(sizeof(formulalist));
	  newlist->next->formula = right2;
	  newlist->next->next = NULL;
	  tnode *left = create_tnode(CONSTANT,CTRUE,s->mdepth,NULL,NULL,NULL);
	  tnode *disj = create_tnode(DISJUNCTION,DISJUNCTION,s->mdepth,NULL,NULL,newlist);
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left,disj,NULL);
	  get_snf_b(set,LITERAL,clause_subtype,mlevel-1,newimp,SYM);

	  /* t_phi => phi : in the predecessor */

	  if (!(is_literal(s->right->left))) {
	    def = new_renaming(s->right->left);
	    if (def == 0) {
	      if (s->right->left->type == NEGATION)  def = s->right->left->left->id;
	      else def = s->right->left->id;
	    }
	    tnode *left3 = create_tnode(PROP,def,s->mdepth,NULL,NULL,NULL);
	    if (!(is_literal(s->right->left))) {
	      tnode *right3 = copy_tree(s->right->left);
	      tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left3,right3,NULL);
	      get_snf_b(set,LITERAL,clause_subtype,mlevel-1,newimpred,SYM);
	    }
	  }
	}
      }
      break;
    }
  }
  return 0;
}


int anchor_snf_b (tnode *root, int set, int mlevel) {
  if (root == NULL) return 0;
  switch (root->type) {
  case SETC:
    {
      create_clauses(set,root->left);
      anchor_snf_b(root->right, root->id, mlevel);
    }
    break;
  case SETF:
    {
      anchor_snf_b(root->left, root->id, mlevel);
      anchor_snf_b(root->right, root->id, mlevel);
    }	
    break;
  case CONJUNCTION:
  case DISJUNCTION:
  case NEGATION:
  case BOX:
  case DIAMOND:
  case PROP:
  case CONSTANT:
  case IMPLICATION:
    {
      //      prop_node *p = insert_pnew_node(numprops,HIGH);
      int def = new_renaming(root);
      tnode *new1 = create_tnode(PROP,def,0,NULL,NULL,NULL);
      tnode *new2 = create_tnode(PROP,def,0,NULL,NULL,NULL);
      
      formulalist *dis = malloc(sizeof(formulalist));
      
      dis->formula = new1;
      dis->next = NULL;

      justification *just = malloc(sizeof(justification));
      just->rule = SNF;
      just->parents = NULL;
      just->literals = NULL;

      int type = LITERAL;
      tnode *left = create_tnode(CONSTANT,CTRUE,0,NULL,NULL,NULL);
      if (ires) {
	type = INITIAL;
	left->id = CSTART;
      }
      
      create_clause_snf_b(set,type,BOX,mlevel,PROP,1,left,dis,just);

      tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,root->mdepth,new2,root,NULL);

      //      printf("\nFormula is:\n");
      //      print_tree(newimp);

      //      printf("\n Starting translation:\n");
      
      get_snf_b(set,LITERAL,BOX,mlevel,newimp,SNF);
      
      free(newimp);
      free(new1);
      free(new2);
      free(dis);
    }
    break;
  }
  return 1;
}
